<?php


namespace Service;


use Doctrine\ORM\EntityManager;
use Exception;
use Teknisa\Model\Teknisa\Entities\CpCall;
use Util\InstanceProvider;
use Util\Utilities;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource;
use Zeedhi\DevOps\Model\Projectmanagement\Repositories\BudgetPlots;
use Zeedhi\DevOps\Util\Utilities as DevOpsUtilities;
use Zeedhi\Framework\ORM\DateTime;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;
use Zeedhi\DevOps\Model\Projectmanagement\Repositories;

class Issue
{
    const ACTIVE = 'A';
    const IS_ISSUE = 'Y';
    const IS_TASK = 'Y';
    const FIRST_COMMENTS = 'Abertura de issue.';
    /**
     * @var EntityManager
     */
    public $entityManager;
    /**
     * @var Utilities
     */
    public $environment;

    public function __construct()
    {
        $this->entityManager  = InstanceProvider::getEntityManager();
        $this->environment    = InstanceProvider::getEnvironment();
    }

    public function createIssueByProductCall($reference, $product, $classification, $description, $board, $reporterName,
                                             $reporterFrom,$sponsorId,$reportedBy)
    {
        $description = $reporterName . " - " . $reporterFrom . PHP_EOL . PHP_EOL . PHP_EOL . $description;

        $now = new \DateTime();
        $dueDate = $now->format('d/m/Y');
        $issueStatus = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskStatus')
            ->findOneBy(array('control' => Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG))->getId();
        $reporterId = InstanceProvider::getEnvironment()->getCurrentUserId();
        $priority = null;
        $percentCompleted = 0;
        $followers = array();
        $labels = null;

        try {
            return $this->saveNewIssue($reference, $classification, $issueStatus, $reporterId, $dueDate, $priority, $percentCompleted, $description,
                $board, $followers, $labels, $sponsorId);
        } catch (Exception $e) {
            throw $e;
        }
    }

    public function findIssueBudgetPlots($issueId,$namePlots)
    {
        return InstanceProvider::getRepository()->getBudget()
            ->findIssueBudgetPlots($issueId,$namePlots);
    }

    public function findIssueBudget($issueId,$nameProposal,$nameDiscount,$namePercent)
    {
        return InstanceProvider::getRepository()->getCpCallLocal()
            ->findIssueBudget($issueId,$nameProposal,$nameDiscount,$namePercent);
    }


    public function saveNewIssue($name, $classification, $issueStatus, $reporter, $dueDate, $priority, $percentCompleted,
                                 $description, $board, $followers, $labels = null, $sponsor = null, $endDate = null,
                                 $productVersionId = null, $version = null, $projectId = null, $versionDev = null,
                                 $productVersionDevId = null, $isBoardIssue = null) {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $name = substr($name, 0, 200);
            $issueStatus 	= $this->getIssueStatus($issueStatus);
            $board 			= $this->entityManager->find(Utilities::ENTITY_DIR.'Project', $board);
            if($endDate){
                $endDate = \DateTime::createFromFormat('d/m/Y', $endDate);
            }
            if($dueDate) {
                $dueDate = \DateTime::createFromFormat('d/m/Y', $dueDate);
            }else {
                $dueDate = new \DateTime();
            }
            if($priority){
                $priority = $this->entityManager->find(Utilities::ENTITY_DIR . 'Priority', $priority);
            }
            /** @var Entities\Classification $classification */
            $classification = InstanceProvider::getRepository()->getClassification()->find($classification);
            $reporter 		= InstanceProvider::getRepository()->getResource()->find($reporter);
            if($isBoardIssue){
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            } else {
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersionDev($projectId, $version, $productVersionId);
            }

            $issue  = InstanceProvider::getFactory()->getTask()->createTask($board, null, $name, $this->environment->getCurrentTime(),
                $dueDate, self::ACTIVE, $description, $percentCompleted, null, null, $issueStatus,
                null, null, self::IS_ISSUE, 'N', $classification, null, $priority, null, null, $endDate, null, null, $productVersion, $productVersionDev);

            $this->entityManager->persist($issue);

            $allocationService = InstanceProvider::getService()->getAllocation();
            if(!empty($reporter)){
                $allocationService->createAllocation($issue, $reporter, Entities\Allocation::TYPE_REPORTER);
            }

            if(!empty($sponsor)){
                $sponsorEntity = $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $sponsor);
                $allocationService->createAllocation($issue, $sponsorEntity, Entities\Allocation::TYPE_SPONSOR);
            }

            $this->setFollowers($followers, $issue);

            if(is_array($labels)){
                $this->setLabels($labels,$issue);
            }
            $this->entityManager->flush();

            $commentType = InstanceProvider::getRepository()->getCommentType()->findOneBy(array('control' => Entities\CommentType::TYPE_OPEN_ISSUE));
            $fistComments = InstanceProvider::getFactory()->getComments()->createComments($issue,self::FIRST_COMMENTS, $commentType);
            $issue->setCurrentTaskStatusIn(Utilities::getEnvironment()->getCurrentTime());
            $this->entityManager->persist($issue);
            $this->entityManager->persist($fistComments);
            $this->viewCommentsSaveData($issue->getId());
            $this->entityManager->flush();

            $reporterAllocation = $allocationService->getAllocationByUserAndTask($reporter->getId(), $issue->getId());
            if(!empty($reporterAllocation)){
                $reporterAllocation->setLastPostView($this->environment->getCurrentTime());
                $this->entityManager->persist($reporterAllocation);
                $this->entityManager->flush();
            }

            $connection->commit();
            self::createCall($name,$reporter,$description,
                $issue->getId(),$issue->getProject()->getId(),$classification->getId());
            try {
                $this->sendOpenIssueMail($issue);
            } catch (Exception $e1) {
            }
            return $issue->getId();
        } catch (Exception $e) {
            try {
                $connection->rollBack();
            } catch (Exception $e1) {
            }
            throw $e;
        }
    }

    private function getIssueStatus($issueStatus)
    {
        $taskStatusRepository = InstanceProvider::getRepository()->getTaskStatus();
        if($issueStatus){
            return $taskStatusRepository->find($issueStatus);
        } else {
            return $taskStatusRepository->findOneBy(array('CONTROL' => Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG));
        }
    }

    public function getProductVersion($productVersionId)
    {
        if($productVersionId){
            return InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
        }
        return null;
    }

    public function getProductVersionDev($projectId, $version, $productVersionId)
    {
        if($productVersionId){
            $productVersionDevId = $this->getProductVersionDevFromProductVersion($version, $projectId);
            if($productVersionDevId){
                return InstanceProvider::getRepository()->getProductVersion()->find($productVersionDevId);
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    public function getProductVersionDevFromProductVersion($version, $projectId)
    {
        $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $projectId));
        $organizationId = InstanceProvider::getEnvironment()->getOrganizationId();
        $versions = explode('.', $version);
        $version = $versions[0];
        $revision = $versions[1];
        $productVersionsLts = InstanceProvider::getRepository()->getProductVersion()->findLtsCurrentProductVersionFromVersion($product->getId(), $organizationId, $version, $revision);
        return !empty($productVersionsLts) ? $productVersionsLts[0]['ID'] : null;
    }

    protected function setFollowers(array $followers, $issue)
    {
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        $allocationRepository = InstanceProvider::getRepository()->getAllocation();
        foreach ($followers as $eachResource) {
            if (!empty($eachResource)) {
                $resourceProject = $resourceRepository->find($eachResource);
                $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), "resource" => $resourceProject->getId(),
                    'type' => Allocation::FOLLOWER));
                if (empty($allocation)) {
                    $allocationService->createAllocation($issue, $resourceProject, Allocation::FOLLOWER);
                }
            }
        }
        $this->entityManager->flush();
    }

    public static function createCall($name,Resource $reportedBy,$description,$issueId,$projectId,$classificationId)
    {
        $call = InstanceProvider::getRepository()->getCpCall()->findOneBy(array('callRelationship'=> $issueId));
        if(!$call){
            $cdOperador = $reportedBy->getExternalCode();
            $call = self::factoryCall($name,$cdOperador,$description,$issueId,$projectId,$classificationId);
            $entityManager = InstanceProvider::getEntityManager();
            $entityManager->persist($call);
            $entityManager->flush();
        }
    }

    public static function factoryCall($reference, $reportedBy, $description, $callRelationship, $product,
                                       $classification)
    {
        $entity = new  CpCall();
        $entity->setReference($reference);
        if (!$reportedBy) {
            $reportedBy = 1;
        }
        $entity->setReportedBy($reportedBy);
        if (!$description) {
            $description = ' ';
        }
        $entity->setDescription($description);
        $entity->setType(CpCall::TYPE_ISSUE);
        $entity->setCallRelationship($callRelationship);
        $entity->setProduct($product);
        $entity->setClassification($classification);
        return $entity;
    }


    public function sendOpenIssueMail(Entities\Task $issue)
    {
        if($issue->getProject()->getNotifyIssueOpening() == Entities\Project::IS_NOTIFY_ISSUE){
            $commenterId = $this->environment->getCurrentUserId();
            $emails    = InstanceProvider::getRepository()->getTask()->getManagersEmailsByIssueId($issue->getId(), $commenterId);
            $issueName = $issue->getName();
            $reporter  = DevOpsUtilities::getResourceNameForSendingMail($issue->getId(), Entities\Allocation::TYPE_REPORTER);
            $sponsor   = DevOpsUtilities::getResourceNameForSendingMail($issue->getId(), Entities\Allocation::TYPE_SPONSOR);
            $organizationName = InstanceProvider::getRepository()->getOrganization()->find($issue->getOrganization())->getName();
            $commenterName = $this->environment->getCurrentUser()->getName();
            $date = $this->environment->getCurrentTime()->format(Utilities::FORMAT_DATE_DEFAULT_NO_SECS);
            $issueTitle  = $issue->getId() . " - " . $issueName;
            $productName = $issue->getProject()->getName();

            foreach ($emails as $email) {
                InstanceProvider::getMailIssueMail()->sendOpenIssueMail($email['EMAIL'], $issueTitle, $commenterName, $date, $issue->getDescription(),
                    $reporter, $sponsor, $issueName, $issue->getId(), $productName, $organizationName);
            }
        }
    }

    public function viewCommentsSaveData($taskId){
        $user = $this->environment->getCurrentUser();
        /** @var Entities\Task $issue */
        $issue = InstanceProvider::getRepository()->getTask()->find($taskId);
        $lastDate = null;
        if($issue){
            /** @var Entities\Allocation $allocation */
            $allocations = InstanceProvider::getRepository()->getAllocation()->findBy(array('task' => $issue ,'resource' => $user));
            if(empty($allocations)){
                $allocations = $this->addFollower($issue, $user);
                $allocations->setLastPostView($this->environment->getCurrentTime());
                $this->entityManager->persist($allocations);
                $this->entityManager->flush();
            }else{
                foreach($allocations as $allocation){
                    $allocation->setLastPostView($this->environment->getCurrentTime());
                    $this->entityManager->persist($allocation);
                    $this->entityManager->flush();
                    $lastDate = $allocation->getLastPostView();
                }
            }
        }
        return $lastDate;
    }

    public function addFollower(Entities\Task $issue, Resource $resource){
        return InstanceProvider::getService()->getAllocation()->createAllocation($issue, $resource, Allocation::FOLLOWER);
    }

    public function findIssue($issueId,$accompanyingIssue = false){
        $issues = InstanceProvider::getRepository()->getTask()->getIssue($issueId, $this->environment->getCurrentUserId());
        foreach($issues as &$issue) {
            $issue['FOLLOWERS'] = Utilities::setSelectMultiInQuery($issue['FOLLOWERS'],true);

            $issue['LABELS'] =  Utilities::setSelectMultiInQuery($issue['LABELS']);

            $issue['CLIENT_ACCEPT'] = Entities\TaskStatus::TASK_STATUS_CONTROL_CLIENT_ACCEPT == $issue['CONTROL'];

            $issue['TASK_CLASSIFICATION_IDS'] =  Utilities::setSelectMultiInQuery($issue['TASK_CLASSIFICATION_IDS']);

            $issue['ACTIONS_OF_EXITS'] = $this->getActionsOfExits($issueId);

            list($issue['LABELS'], $issue['LABELS_NOT_EDITABLE']) =
                $this->getLabelsForEditAndNotEditInManagerIssue($issue['LABELS'], $issue['BOARD'],$accompanyingIssue);

            $issue['ACTIONS_REDO_BUDGET'] = $this->allowRedoBudget($issueId);
        }
        return $issues;
    }


    public function allowRedoBudget($issueId){
        $return = false;
        $budgetRepository = InstanceProvider::getRepository()->getBudget();
        /** @var Entities\Budget $budget */
        $budget = $budgetRepository->findOneBy(array('task' => $issueId));
        if($budget && !$budget->getExternalId()){
            $organizationId = \Util\Utilities::getEnvironment()->getCurrentOrganizationId();
            $urlSm = $budgetRepository->getUrlIntegrationSM($organizationId);
            if($urlSm &&
                $budget->getTask()->getTaskStatus() &&
                $budget->getTask()->getTaskStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG){

                /** @var Entities\Comments $comment */
                $comment = InstanceProvider::getRepository()->getComments()->findOneBy(array('task' => $issueId ,'commentType' => 3),array('createdAt'=> 'DESC'));
                if($comment){
                    $classificationId = $budget->getTask()->getClassification()->getId();
                    $taskSprint = InstanceProvider::getRepository()->getTask()->getTaskRelationshipByIssue($issueId,  " AND CLASSIFICATION_ID = $classificationId AND PERCENT <> 100");
                    if(!empty($taskSprint)){
                        $return = true;
                    }
                }
            }
        }
        return $return;
    }

    /**
     * @return array
     */
    public function getImInMeetingProjectIds()
    {
        $projectType = Utilities::getProjectTypeMeeting();
        /** @var Entities\ResourceProject[] $projectResources */
        $meetingProject = InstanceProvider::getRepository()->getProject()->findBy(array('projectType' => $projectType));;
        $projectResources = InstanceProvider::getRepository()->getResourceProject()->findBy(array(
            'resourceC' => $this->environment->getCurrentUserId(),
            'project' => $meetingProject));
        $projectIds = Utilities::getArrayByArrayObjectWithGet($projectResources,array('getProject','getId'));

        return $projectIds;
    }

    public function getLabelsForEditAndNotEditInManagerIssue(array $labelIds,$board = null, $accompanyingIssue = false, $editable = array(),
                                                             $notEditable = array())
    {
        $meetingProjectIds = $this->getImInMeetingProjectIds();
        if($board && !$accompanyingIssue){
            $meetingProjectIds[] = $board;
        }
        $userId = $this->environment->getCurrentUserId();
        foreach ($labelIds as $labelId) {
            /** @var Entities\Label $label */
            $label = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Label')->find($labelId);
            $relatedResource = $label->getResource();
            if ($relatedResource) {
                if ($relatedResource->getId() == $userId) {
                    $editable[] = $labelId;
                } else {
                    $notEditable[] = $labelId;
                }
            } else if ($label->getProject() && in_array($label->getProject()->getId(), $meetingProjectIds)) {
                $editable[] = $labelId;
            } else {
                $notEditable[] = $labelId;
            }
        }
        return array($editable,$notEditable);
    }

    public function getActionsOfExits($issueId): array
    {
        $actionsOfExits = InstanceProvider::getRepository()->getTask()->getActionsOfExits($issueId);

        /** @var Entities\Task $issue */
        $issue = InstanceProvider::getRepository()->getTask()->find($issueId);

        foreach ($actionsOfExits as &$actionsOfExit) {
            $actionsOfExit['COMMENT_DESCRIPTION'] = $actionsOfExit['COMMENT_DESCRIPTION']?$actionsOfExit['COMMENT_DESCRIPTION']:'';
        }

        if ($issue->getTaskStatus()->getControl() === 'H') {
            $actionsOfExits = array_merge($actionsOfExits, $this->getActionsFinishOrDecline());
        }

        return $actionsOfExits;
    }

    /**
     * @return array[]
     */
    private function getActionsFinishOrDecline()
    {
        /** @var Entities\CommentType $commentTypeRefuse */
        $commentTypeRefuse = InstanceProvider::getRepository()->getCommentType()->findOneBy(array('control'=>'R'));
        /** @var Entities\TaskStatus $taskStatusRefuse */
        $taskStatusRefuse = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('control' => Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG));
        return array(
            array(
                'ID' => 'FINALIZE',
                'ACTION_LABEL' => 'Finalizar Issue',
                'TASK_TYPE_TARGET' => null,
                'COMMENT_TYPE_ID' => null,
                'ATTACHMENT_REQUIRED' => 'N',
                'COMMENT_DESCRIPTION' => '',
                'SPECIAL_FUNCTION' => 'F',
                'HOMOLOGATED' => 'Y'
            ),
            array(
                'ID' => 'REFUSE',
                'ACTION_LABEL' => 'Recusar Issue',
                'TASK_TYPE_TARGET' => $taskStatusRefuse->getTaskType()->getId(),
                'COMMENT_TYPE_ID' => $commentTypeRefuse->getId(),
                'ATTACHMENT_REQUIRED' => 'N',
                'SPECIAL_FUNCTION' => 'R',
                'COMMENT_DESCRIPTION' => null,
                'HOMOLOGATED' => 'Y'
            ),
        );
    }

    /**
     * @param $issueId
     * @param $reference
     * @param $product
     * @param $classification
     * @param $description
     * @param $board
     * @param Resource $resource
     * @param $resourceSponsorId
     * @return mixed
     * @throws Exception
     */
    public function updateIssueByProductCall($issueId, $reference, $product, $classification, $description, $board, Resource $resource, $resourceSponsorId) {

        $now = new \DateTime();
        $dueDate = $now->format('d/m/Y');
        $issue              = $this->findIssue($issueId)[0];
        $issueStatus        = $issue['ISSUE_STATUS'];
        $reporter           = $resource->getId();
        $priority           = $issue['PRIORITY'];
        $percentCompleted   = $issue['PERCENT_COMPLETED'];
        $followers          = $issue['FOLLOWERS'];
        $sponsor            = $resourceSponsorId;
        $labels             = $issue['LABELS'];
        try{
            $issueId = $this->saveIssue($issueId, $reference, $classification, $issueStatus, $reporter, $dueDate, $priority, $percentCompleted,
                $description, $followers, null, $board, $labels, $sponsor);

              return $this->findIssue($issueId['ID'])[0];
        } catch(Exception $e){
            throw $e;
        }
    }

    public function saveIssue(
        $id, $name, $classification, $issueStatus, $reporter, $dueDate, $priority,
        $percentCompleted, $description, $followers, $endDate, $board = null,
        $labels = null, $sponsor = null, $labelsNotEditable = array(), $isRelevant = Entities\Allocation::NOT_RELEVANT,
        $productVersionId = null, $version = null, $projectId = null, $isBoardIssue = null, $productVersionDevId = null, $issueResponse = null
    ){

        $connection = $this->entityManager->getConnection();
        try{
            $connection->beginTransaction();

            $name = substr($name, 0, 200);
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            $issue          = InstanceProvider::getRepository()->getTask()->find($id);
            $issueStatus    = $this->entityManager->find(Utilities::ENTITY_DIR . 'TaskStatus', $issueStatus);
            $reporter       = empty($reporter) ? null : $resourceRepository->find($reporter);
            $dueDate        = empty($dueDate) ?  $this->environment->getCurrentTime() : \DateTime::createFromFormat('d/m/Y', $dueDate);
            $endDate        = empty($endDate) ?  null : \DateTime::createFromFormat('d/m/Y', $endDate);
            $classification = $this->entityManager->find(Utilities::ENTITY_DIR . 'Classification', $classification);

            if($priority){
                $priority = $this->entityManager->find(Utilities::ENTITY_DIR . 'Priority', $priority);
            }

            $frontReload = false;
            if($issue->getName() != $name || $issue->getDescription() != $description ||
                $issue->getTaskStatus()->getId() != $issueStatus->getId() ||
                $issue->getClassification()->getId() != $classification->getId()){
                $frontReload = true;
            }

            if($issue->getProductVersionDev()){
                if($issue->getProductVersionDev()->getId() != $productVersionDevId){
                    $frontReload = true;
                }
            } else if($productVersionDevId){
                $frontReload = true;
            }


            if($priority && $issue->getPriority()) {
                if($priority->getId() != $issue->getPriority()->getId()){
                    $frontReload = true;
                }
            } else if($priority && !$issue->getPriority()) {
                $frontReload = true;
            } else if(!$priority && $issue->getPriority()) {
                $frontReload = true;
            }

            if($board){
                $board = $this->entityManager->find(Utilities::ENTITY_DIR . 'Project', $board);
            }else{
                $board = $issue->getProject();
            }
            if($isBoardIssue){
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            } else {
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            }
            $issue = InstanceProvider::getFactory()->getTask()->updateTask($issue, $board, null, $name, $this->environment->getCurrentTime(), $dueDate, self::ACTIVE, $description, $percentCompleted, null, null, $issueStatus, null, null, 'N', self::IS_ISSUE, $classification,null,	$priority, null, null, null, $productVersion, $productVersionDev, $issueResponse);

            $issue->setEndDate($endDate);
            $this->entityManager->persist($issue);
            $this->setLabelsIssue($labels, $labelsNotEditable, $issue);
            $allocationService = InstanceProvider::getService()->getAllocation();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
            if(!empty($reporter)){
                /** @var Entities\Allocation $allocation */
                $allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::REPORTER));
                if(empty($allocation)){
                    $allocationService->createAllocation($issue, $reporter, Allocation::REPORTER, null, true);
                }
                else{
                    $allocationService->updateAllocation($allocation, $issue, $reporter, Allocation::REPORTER, null, true);
                }
            }
            if(!empty($sponsor)){
                $sponsorEntity	= $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $sponsor);
                $allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::SPONSOR));
                if(empty($allocation)){
                    $allocationService->createAllocation($issue, $sponsorEntity, Allocation::SPONSOR, null, true);
                } else {
                    $allocationService->updateAllocation($allocation, $issue, $sponsorEntity, Allocation::SPONSOR, null, true);
                }
            }

            foreach ($followers as $eachResource) {
                if(!empty($eachResource)){
                    $resourceProject = $resourceRepository->find($eachResource);
                    $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), "resource" => $resourceProject->getId(),
                        'type' => Allocation::FOLLOWER));
                    if (empty($allocation)) {
                        $allocationService->createAllocation($issue, $resourceProject, Allocation::FOLLOWER, null, true);
                    }
                }
            }
            /** @var Entities\Allocation[] $allocation */
            $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), 'type' => Allocation::FOLLOWER));

            foreach ($allocation as $eachAllocation) {
                if(!in_array($eachAllocation->getResource()->getId(), $followers)){
                    $this->entityManager->remove($eachAllocation);
                }
            }

            $user = $this->environment->getCurrentUserId();
            if(!empty($user)){
                $resourceEntity	= $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $user);
                $allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::FOLLOWER, 'resource' => $user));
                if($allocation){
                    $allocationService->updateAllocation($allocation, $issue, $resourceEntity, Allocation::FOLLOWER, null, true, $isRelevant);
                } else if($isRelevant ==  Entities\Allocation::RELEVANT) {
                    $allocationService->createAllocation($issue, $resourceEntity, Allocation::FOLLOWER, null, true, $isRelevant);
                }
            }

            $this->entityManager->flush();
            $connection->commit();


            return array('ID' => $issue->getId(), 'RELOAD' => $frontReload);
        }
        catch(Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $labels
     * @param $labelsNotEditable
     * @param $issue
     */
    protected function setLabelsIssue($labels, $labelsNotEditable, $issue)
    {
        $labels = array_unique(array_merge($labelsNotEditable, $labels));
        $this->setLabels($labels, $issue);
    }

    public function setLabels(array $labels,Entities\Task $task)
    {
        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        $labelRepository = InstanceProvider::getRepository()->getLabel();
        $taskLabelRepository = InstanceProvider::getRepository()->getTaskLabel();
        foreach ($labels as $labelId) {
            if(!empty($labelId)){
                /** @var  Entities\Label $label */
                $label = $labelRepository->find($labelId);
                $taskLabel = $taskLabelRepository	->findBy(array("task" => $task->getId(),
                    "label" => $label->getId()));
                if (empty($taskLabel)) {
                    $taskLabel = $taskLabelFactory->createTaskLabel($task, $label);
                    $this->entityManager->persist($taskLabel);
                }
            }
        }
        $this->entityManager->flush();
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskLabel[] $taskLabels */
        $taskLabels = $taskLabelRepository->findBy(array("task" => $task->getId()));
        foreach ($taskLabels as $taskLabel) {
            if(!in_array($taskLabel->getLabel()->getId(), $labels)){
                $this->entityManager->remove($taskLabel);
                $this->entityManager->flush();
            }
        }
    }

    public function validateEditBudget($issueId){

        $validate = $this->allowRedoBudget($issueId);
        /** @var Entities\Task $issue */
        $issue = InstanceProvider::getRepository()->getTask()->find($issueId);
        if($issue->getTaskStatus()->getId() === 32){
            $validate = true;
        }
        return $validate;
    }

    public function issueBudgetSave($data){
        if(isset($data[0]['changed']) && $data[0]['changed']){
            $rows = [];
            foreach ($data as $datum) {
                if(!empty($datum['ISSUE_ID'])){
                    unset($datum['groupHeaders']);
                    $rows[] = $datum;
                }
            }
            $this->issueBudgetsaveItens($rows);
        }
    }

    public function saveMaintenanceCost($row)
    {
        $entityManager = InstanceProvider::getEntityManager();
        $connection = $entityManager->getConnection();
        try {
            $connection->beginTransaction();
            if ($row['ISSUE_ID'] && !isset($row['changed']) && $row['changed']) {
                $issue = InstanceProvider::getRepository()->getTask()->find($row['ISSUE_ID']);
                /** @var Entities\Budget $budget */
                $budget = InstanceProvider::getRepository()->getBudget()->findOneBy(array('task' => $issue));

                $valueDiscount = str_replace(",", ".", $row['VALUE_DISCOUNT']);
                $valueManut = str_replace(",", ".", $row['MANUT_VALUE']);
                $valueDiscount = floatval(preg_replace('/[^\d\.]+/', '', $valueDiscount));
                $valuePercent = str_replace(",", ".", $row['PERCENTAGE_DISCOUNT']);
                if($valueManut < $valueDiscount){
                    $valueDiscount = $valueManut;
                }
                if($valuePercent > 100){
                    $valuePercent = 100;
                }
                $budget->setMaintenancePercent($valuePercent);
                $budget->setMaintenanceValue($valueDiscount);

                $entityManager->flush();
                $connection->commit();
            }
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    public function findMaintenanceCost($issueId)
    {
        $budgetRepository = InstanceProvider::getRepository()->getBudget();
        $organizationId = InstanceProvider::getEnvironment()->getCurrentOrganizationId();
        $maintenance = $budgetRepository->MaintenanceCost($organizationId);
        $result = array();
        if($maintenance){
            $result['MAINTENANCE_PERCENTAGE'] = $maintenance;
            $budgetPrice = $budgetRepository->findManutIssueBudget($issueId);
            $finalPrice = round($budgetPrice[0]["TOTAL_VALUE"], 2);
            $maintenancePrice = round($finalPrice * $maintenance / 100, 2);
            $maintenancePrice = number_format($maintenancePrice, 2, '.', '');

            $maintenanceDiscounts = $budgetRepository->findManutIssueDiscount($issueId);
            if($maintenanceDiscounts && $maintenanceDiscounts[0]){
                $result['VALUE_DISCOUNT'] = $maintenanceDiscounts[0]['MAINTENANCE_VALUE'];
                $result['PERCENTAGE_DISCOUNT'] =  $maintenanceDiscounts[0]['MAINTENANCE_PERCENT'];
                $result['TOTAL_VALUE'] = $maintenancePrice - $maintenanceDiscounts[0]['MAINTENANCE_VALUE'];
            }

            $result['MANUT_VALUE'] = $maintenancePrice;
        }else{
            $result['MANUT_VALUE'] = 0;
            $result['TOTAL_VALUE'] = 0;
            $result['PERCENTAGE_DISCOUNT'] = 0;
            $result['VALUE_DISCOUNT'] = 0;
            $result['MAINTENANCE_PERCENTAGE'] = 2;
        }
        $result['VALIDATE_EDIT']  = $this->validateEditBudget($issueId);
        return $result;
    }

    public function issueBudgetPlotsSave($rows){

        $connection = InstanceProvider::getEntityManager()->getConnection();
        try {
            $connection->beginTransaction();

            $instanceRepository = InstanceProvider::getRepository();
            $entityManager = InstanceProvider::getEntityManager();
            $factoryBudgetPlots = InstanceProvider::getFactory()->getBudgetPlots();
            /** @var BudgetPlots $repositoryBudgetPlots */
            $repositoryBudgetPlots = $instanceRepository->getBudgetPlots();

            /** @var Entities\Budget $budget */
            $budget = $instanceRepository->getBudget()->find($rows[0]['BUDGET_ID']);
            /** @var Entities\BudgetPlots[] $budgetPlots */
            //$budgetPlots = $instanceRepository->getBudgetPlots()->findby(array('budget' => $budget),array('orderPlots' => 'ASC'));
            $lastParcel = 0;

            foreach ($rows as $row) {
                $orderPlots = intval($row['ORDER_PLOTS_NUM']);
                if($lastParcel < $orderPlots){
                    $lastParcel = $orderPlots;
                }
                $valuePlots = $row['VALUE_PLOTS'];
                $daysPayment = $row['DAYS_PAYMENT'];
                /** @var Entities\BudgetPlots $budgetPlots */
                $budgetPlots = $repositoryBudgetPlots->findOneBy(array('budget' => $budget,'orderPlots' => $orderPlots));
                if($budgetPlots){
                    $budgetPlots = $factoryBudgetPlots->save($valuePlots,$daysPayment,$budgetPlots);
                }else{
                    $budgetPlots = $factoryBudgetPlots->save($valuePlots,$daysPayment,null,$budget,$orderPlots);
                }
                $entityManager->persist($budgetPlots);
            }
            $entityManager->flush();

            if($lastParcel){
                $budgetPlotsIdToDelete = $repositoryBudgetPlots->findBudgetPlotsToDelete($budget->getId(),$lastParcel);
                $budgetPlotsToDelete = $repositoryBudgetPlots->findBy(array('id' => $budgetPlotsIdToDelete));
                foreach ($budgetPlotsToDelete as $parcelToDelete) {
                    $entityManager->remove($parcelToDelete);
                }
                if(!empty($budgetPlotsToDelete)){
                    $entityManager->flush();
                }
            }
            $connection->commit();
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    public function issueBudgetRowSave($issueId,$budgetItemId,$proposalQuantity,$proposalValue)
    {

        $entityManager = InstanceProvider::getEntityManager();
        $instanceRepository = InstanceProvider::getRepository();
        $instanceFactory = InstanceProvider::getFactory();
        /** @var Entities\Task $issue */
        $issue = $instanceRepository->getTask()->find($issueId);
        /** @var Entities\Budget $budget */
        $budget = $instanceRepository->getBudget()->findOneBy(array('task' => $issue));
        $budget = $instanceFactory->getBudget()->save($budget,$issue);
        if($budgetItemId == '999999XXXX'){
            $proposalQuantity = abs($proposalQuantity);
            if($proposalQuantity > 100){
                $proposalQuantity = 100;
            }
            $budget->setDiscountPercent($proposalQuantity);
            $budget->setDiscountValue($proposalValue);
            $entityManager->persist($budget);
            $entityManager->flush();
        }else{
            /** @var Entities\BudgetItem $budgetItem */
            $budgetItem = $instanceRepository->getBudgetItem()->find($budgetItemId);
            /** @var Entities\BudgetItemBudget $budgetItemBudget */
            $budgetItemBudget = $instanceRepository->getBudgetItemBudget()
                ->findOneBy(array( 'budgetItem' => $budgetItem, 'budget' => $budget, ));
            if($proposalQuantity){
                $valueItem = $budgetItem->getValue();
                $unitMeasure = $budgetItem->getUnitMeasure();
                $budgetItemBudget = $instanceFactory->getBudgetItemBudget()->save($budgetItemBudget,$budget,$budgetItem,
                    $proposalQuantity,$valueItem,$unitMeasure);
                $entityManager->persist($budgetItemBudget);
                $entityManager->flush();
            }else if($budgetItemBudget){
                $entityManager->remove($budgetItemBudget);
                $entityManager->flush();
            }
        }
        return $budget->getId();
    }

    public function issueBudgetsaveItens($rows){

        $entityManager = InstanceProvider::getEntityManager();
        $instanceRepository = InstanceProvider::getRepository();
        $connection = $entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $budgetId = null;
            foreach ($rows as $row) {
                $issueId = $row['ISSUE_ID'];
                $budgetItemId = $row['BUDGET_ITEM_ID'];
                $proposalQuantity = $row['PROPOSAL_QUANTITY'];
                $proposalValue = $row['PROPOSAL_VALUE'];
                $budgetId = $this->issueBudgetRowSave($issueId, $budgetItemId,
                    $proposalQuantity, $proposalValue);
            }

            $budget = InstanceProvider::getRepository()->getBudget()->findOneBy(array('task' => $issueId));

            if($budget && $budget->getMaintenancegetPercent()){
                $budgetRepository = InstanceProvider::getRepository()->getBudget();
                $organizationId = InstanceProvider::getEnvironment()->getCurrentOrganizationId();

                $budgetPrice = $budgetRepository->findManutIssueBudget($issueId);
                $maintenance = $budgetRepository->MaintenanceCost($organizationId);
                $finalPrice = round($budgetPrice[0]["TOTAL_VALUE"] , 2);
                $maintenancePrice = round($finalPrice * $maintenance / 100, 2);
                $maintenancePrice = number_format($maintenancePrice, 2);

                if($maintenancePrice > 0){
                    $discountValue = $budget->getMaintenancegetPercent() * $maintenancePrice / 100;

                    $budget->setMaintenanceValue($discountValue);

                }else{
                    $budget->setMaintenanceValue(0);
                }
                $entityManager->flush();
            }
            $repositoryBudgetPlots = $instanceRepository->getBudgetPlots();
            $budgetPlotsToDelete = $repositoryBudgetPlots->findBy(array('budget' => $budgetId));
            foreach ($budgetPlotsToDelete as $parcelToDelete) {
                $entityManager->remove($parcelToDelete);
            }
            if(!empty($budgetPlotsToDelete)){
                $entityManager->flush();
            }
            $connection->commit();
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }

    public function closeIssue($issueIds){
        $user = $this->environment->getCurrentUser();
        /**@var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties $taskProperties*/
        $taskProperties = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties')->findOneBy(array('organization' =>$this->environment->getCurrentOrganizationId()));
        if($taskProperties){
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
            foreach($issueIds AS $issueId){
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $issue*/
                $issue = $taskRepository->find($issueId);
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation $reporter */
                $reporter = $allocationRepository->findOneBy(array('type'=>\Service\Allocation::REPORTER,'resource' => $user,'task'=>$issue));
                if(empty($reporter)){
                    throw new \Exception('Essa issue só pode ser finalizada pelo solicitante!');
                }
                $issue->setPercentCompleted(100);
                $issue->setTaskStatus($taskProperties->getFinalTaskStatus());
                $issue->setCurrentTaskStatusIn((new \DateTime()));
                $this->entityManager->persist($issue);
            }
            $this->entityManager->flush();
        }else{
            throw new \Exception('Padrão de tarefa não cadastrada!');
        }
        return $issue;
    }

    public function getClientAcceptTask(Entities\Task $issue){
        $tasks = $issue->getTaskForIssue();
        foreach ($tasks as $task) {
            if($task->getPercentCompleted() != 100 &&
                $task->getClassification() &&
                $task->getClassification()->getControl() == Entities\Classification::CONTROL_CLIENT_ACCEPT )
            {
                return $task;
            }
        }
        return null;
    }

    public function updateTaskAndAllocationStatus(Entities\Task $task, Entities\TaskStatus $taskStatus)
    {
        $task->setTaskStatus($taskStatus);
        $this->entityManager->persist($task);
        $allocations = InstanceProvider::getRepository()->getAllocation()->findBy(array('task' => $task));
        /** @var Entities\Allocation $allocation */
        foreach($allocations as $allocation) {
            $allocation->setTaskStatus($taskStatus);
            $this->entityManager->persist($allocation);
        }
    }

    public function clientRefuse($issueId, $commentTypeId, $description, $document)
    {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        $taskRepository = InstanceProvider::getRepository()->getTask();
        try {
            /** @var Entities\Task $issue */
            $issue = $taskRepository->find($issueId);

            $tasksFromIssue = $taskRepository->getTasksFromIssue($issueId);
            foreach ($tasksFromIssue as $taskFromIssue) {
                $taskFromIssue = $taskRepository->find($taskFromIssue['ID']);
                $taskFromIssue->setUsedInFraming('N');
                $this->entityManager->persist($taskFromIssue);
            }
            $task = $this->getClientAcceptTask($issue);
            $task->setPercentCompleted(100);
            $this->updateTaskAndAllocationStatus($task, Utilities::getTaskStatusIsConcluded());
            if($commentTypeId){
                $commentType = InstanceProvider::getRepository()->getCommentType()->find($commentTypeId);
            }else{
                $commentType = InstanceProvider::getRepository()->getCommentType()->findOneBy(array('CONTROL'=>'R'));
            }
            $comment = InstanceProvider::getFactory()->getComments()->createComments($issue, $description, $commentType);
            $issue->setCurrentTaskStatusIn(Utilities::getEnvironment()->getCurrentTime());
            /** @var Entities\TaskStatus $taskStatus */
            $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('control' => Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG));
            $issue->setTaskStatus($taskStatus);
            $this->entityManager->persist($issue);
            $this->entityManager->persist($task);
            $this->entityManager->persist($comment);
            $this->entityManager->flush();

            Integration\IntegrationSM::integrationSMCancel($issueId, $issue->getOrganization());
            /** @var Entities\Budget $budget */
            $budget = InstanceProvider::getRepository()->getBudget()->findOneBy(array('task' => $issueId));
            if($budget){
                $budget->setExternalId(null);
                $this->entityManager->persist($budget);
                $this->entityManager->flush();
            }

            if ($document) {
                Utilities::getAttachmentService()->addDocumentComment($comment->getId(), $document);
            }
            Utilities::getCommentService()->sendsMailAdvisingAboutNewIssueComment($issue->getId(), $description);

            $connection->commit();
            return $issue;
        } catch (Exception $e) {
            try {
                $connection->rollBack();
            } catch (Exception $ex) {}
            throw $e;
        }
    }

    public function sendOutComment($issueId, $document, $taskTypeTargetId, $commentTypeId, $description,$specialFunction = null){
        /** @var Entities\Task $issue */
        $issue = InstanceProvider::getRepository()->getTask()->find($issueId);
        $taskRepository = InstanceProvider::getRepository()->getTask();
        /** @var Entities\TaskType $taskTypeTarget */
        $taskTypeTarget = InstanceProvider::getRepository()->getTaskType()->find($taskTypeTargetId);
        $taskStatus = $taskTypeTarget->getInitialTaskStatus();
        if(!$taskStatus){
            $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('TASK_TYPE' => $taskTypeTargetId));
        }
        $issue->setTaskStatus($taskStatus);
        $issue->setCurrentTaskStatusIn(new \DateTime());
        $this->entityManager->persist($issue);
        if($specialFunction == 'R'){
            $tasksFromIssue = $taskRepository->getTasksFromIssue($issueId);
            foreach ($tasksFromIssue as $taskFromIssue) {
                $taskFromIssue = $taskRepository->find($taskFromIssue['ID']);
                $taskFromIssue->setUsedInFraming('N');
                $this->entityManager->persist($taskFromIssue);
            }
        }
        /** @var Entities\CommentType $commentType */
        $commentType = InstanceProvider::getRepository()->getCommentType()->find($commentTypeId);
        $comment = InstanceProvider::getFactory()->getComments()->createComments($issue, $description,$commentType);
        $this->entityManager->persist($comment);
        $this->entityManager->flush();
        if($document){
            Utilities::getAttachmentService()->addDocumentComment($comment->getId(), $document);
        }
        if($taskStatus->getControl() == 'H'){
            $this->sendMailOnAvailableIssue($issue->getId(), $description);
        }else{
            Utilities::getCommentService()->sendsMailAdvisingAboutNewIssueComment($issue->getId(), $description);
        }
    }

    public function sendMailOnAvailableIssue($issueId, $description = null){
        $requestHTMLBuild = true;
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $resourceId    = $this->environment->getCurrentUserId();
        $emails         = $taskRepository->getRelatedResourcesEmailsByIssueId($issueId, $resourceId);
        $issue          = $taskRepository->find($issueId);
        $issueName      = $issue->getName();
        $reporter       = DevOpsUtilities::getResourceNameForSendingMail($issueId, Entities\Allocation::TYPE_REPORTER);
        $sponsor        = DevOpsUtilities::getResourceNameForSendingMail($issueId, Entities\Allocation::TYPE_SPONSOR);
        $organizationName = InstanceProvider::getRepository()->getOrganization()->find($issue->getOrganization())->getName();
        $commenterName  = $this->environment->getCurrentUser()->getName();
        $issueTitle     = $issueId." - ".$issueName;
        $taskIds        = $taskRepository->getRelationedTasks($issueId);
        $date           = $this->environment->getCurrentTime()->format(Utilities::FORMAT_DATE_DEFAULT_NO_SECS);
        $productName    = $issue->getProject()->getName();
        $mailMessage = '';
        $additionalMessage = '(sem necessidade de implementação)';

        foreach ($taskIds as $taskId) {
            /** @var Entities\Task $task */
            $task = $taskRepository->find($taskId['ID']);
            if($task){
                $mailMessage .= InstanceProvider::getMailIssueMail()->buildBodyMessage( $task);
            }
        }
        if(!empty($emails)){
            foreach($emails as $email){
                InstanceProvider::getMailIssueMail()->sendAvailableMail($email['EMAIL'],$issueTitle , $commenterName, $date, $mailMessage,
                    '.', $reporter, $sponsor, $issueName, $issueId, $productName, $description, $organizationName, $requestHTMLBuild, null, false, false, null, $additionalMessage);
            }
        }
    }
}
