import { AlertServiceCustom } from '@/services/AlertServiceCustom';
import { DialogServiceCustom } from '@/services/DialogServiceCustom';
import { InstanceService } from '@/services/InstanceService';
import { DialogService, GridEditable, LoadingService, Date, ModalService, Modal } from '@zeedhi/common';
import { Http, Metadata } from '@zeedhi/core';

export class ValoresOrcaCliController {
    public currentRow: any = {};

    private modalValoresOrcaClie!: Modal;

    public async abrirModalValoresOrcaCli({ row }: any) {
        LoadingService.show();
        this.modalValoresOrcaClie = await ModalService.createFromJsonPost('', { jsonPath: 'valoresOrcaCli/modal/valoresOrcaCliModal' });
        await this.modalValoresOrcaClie.show();
        this.currentRow = row;
        try {
            const response = await Http.post('/qryVigencia', {
                CDCLIENTE: this.currentRow.CDCLIENTE,
            });
            let dtinivaitorc = '';
            if (response.data.data.length === 0) {
                Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value = dtinivaitorc;
            } else {
                dtinivaitorc = response.data.data[0].DTINIVAITORC;
                Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value = dtinivaitorc;
            }
            const gridItensOrcamento = InstanceService.getInstanceTekGrid('gridItensOrcamento');
            gridItensOrcamento.editing = true;
            gridItensOrcamento.datasource.setFilter({
                CDCLIENTE: this.currentRow.CDCLIENTE,
                DTINIVAITORC: dtinivaitorc,
            });
        } catch (e) {
            LoadingService.hide();
            DialogServiceCustom.showErrorBackend(e);
        }
        LoadingService.hide();
    }

    public closeModalValoresOrcaCliModal() {
        this.modalValoresOrcaClie.hide();
        this.modalValoresOrcaClie.destroy();
    }

    public dialogExcluirValoresOrcaCli() {
        DialogService.show({
            name: 'dialog-excluir-valores',
            type: 'warning',
            title: 'Atenção',
            text: 'Deseja realmente excluir os valores de orçamento dos registros selecionados?',
            buttons: [
                {
                    name: 'NAO',
                    component: 'ZdButton',
                    label: 'Não',
                    outline: true,
                    events: {
                        click: () => {
                            DialogService.hide();
                        },
                    },
                },
                {
                    name: 'SIM',
                    component: 'ZdButton',
                    label: 'Sim',
                    events: {
                        click: () => {
                            this.excluirValoresOrcaCli();
                            DialogService.hide();
                        },
                    },
                },
            ],
        });
    }

    public async excluirValoresOrcaCli() {
        const gridValoresOrcaCli = InstanceService.getInstanceTekGrid('gridValoresOrcaCli');
        LoadingService.show();
        try {
            await Http.post('/qryDelItemOrca', {
                ROWS: gridValoresOrcaCli.selectedRows,
            });
            await gridValoresOrcaCli.datasource.get();
            gridValoresOrcaCli.selectedRows = [];
        } catch (e) {
            LoadingService.hide();
            DialogServiceCustom.showErrorBackend(e);
        }
        LoadingService.hide();
    }

    public async saveValoresOrcaCliTab() {
        const formValoresOrcaCliModal = InstanceService.getInstanceZdForm('formValoresOrcaCliModal');
        const checkValidation = formValoresOrcaCliModal.validate();
        if (checkValidation) {
            const gridItens = Metadata.getInstance<GridEditable>('gridItensOrcamento');
            const editedRows = gridItens.getEditedRows();
            if (editedRows.length > 0) {
                const arrayItens: any = [];
                editedRows.forEach((row) => {
                    arrayItens.push(row);
                });
                LoadingService.show();
                try {
                    await Http.post('/saveValoresOrcaCliTab', {
                        DTINIVAITORC: Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value,
                        CDCLIENTE: this.currentRow.CDCLIENTE,
                        ITENS: arrayItens,
                    });
                    const response = await Http.post('/qryVigencia', {
                        CDCLIENTE: this.currentRow.CDCLIENTE,
                    });
                    let dtinivaitorc = '';
                    if (response.data.data.length === 0) {
                        Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value = dtinivaitorc;
                    } else {
                        dtinivaitorc = response.data.data[0].DTINIVAITORC;
                        Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value = dtinivaitorc;
                    }
                    AlertServiceCustom.defaultSaveSuccessAlert();
                    await gridItens.datasource.get();
                    await gridItens.cancelEditedRows();
                } catch (e) {
                    LoadingService.hide();
                    DialogServiceCustom.showErrorBackend(e);
                }
                LoadingService.hide();
            }
        } else {
            DialogServiceCustom.messageErrorValidate();
        }
    }

    public realoadGridItens() {
        const gridItensOrcamento = InstanceService.getInstanceTekGrid('gridItensOrcamento');
        gridItensOrcamento.datasource.setFilter({
            CDCLIENTE: this.currentRow.CDCLIENTE,
            DTINIVAITORC: Metadata.getInstance<Date>('formValoresOrcaCliModal-DTINIVAITORC').value,
        });
    }

    public async saveNrParcelaValoresOrcaClie() {
        const gridValoresOrcaCli = InstanceService.getInstanceTekGrid('gridValoresOrcaCli');
        const editedRows = gridValoresOrcaCli.getEditedRows();
        LoadingService.show();
        try {
            await Http.post('/saveNrParcelaValoresOrcaClie', {
                ROWS: editedRows,
            });
            AlertServiceCustom.defaultSaveSuccessAlert();
            await gridValoresOrcaCli.datasource.get();
        } catch (e) {
            LoadingService.hide();
            DialogServiceCustom.showErrorBackend(e);
        }
        LoadingService.hide();
    }
}
